"""
AttriBox implements a lazily instantiated and strongly typed descriptor
class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.core import Object
from worktoy.core.sentinels import DELETED

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Self


class AttriBox(Object):
  """
  AttriBox implements a lazily instantiated and strongly typed descriptor
  class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __field_type__ = None  # The type of the objects stored in the box.

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def getFieldType(self) -> type:
    return self.__field_type__

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _createFieldObject(self, ) -> Any:
    """
    Creates a new instance of the field type. This method is called when
    the field is accessed for the first time.
    """
    instance = self.getContextInstance()
    fieldType = self.getFieldType()
    owner = type(instance)
    args = self.getPosArgs(THIS=instance, OWNER=owner, DESC=self)
    kwargs = self.getKeyArgs(THIS=instance, OWNER=owner, DESC=self)
    return fieldType(*args, **kwargs)

  def __instance_get__(self, *args, **kwargs) -> Any:
    """
    Returns the value of the field for the given instance. If the value is
    not set, it initializes it with a new instance of the field type.
    """
    instance = self.getContextInstance()
    pvtName = self.getPrivateName()
    if hasattr(instance, pvtName):
      return getattr(instance, pvtName)
    fieldObject = self._createFieldObject()
    setattr(instance, pvtName, fieldObject)
    return self.__instance_get__(instance, _recursion=True)

  def __instance_set__(self, value: Any, *args, **kwargs) -> None:
    """
    Sets the value of the field for the given instance. If the value is
    not set, it initializes it with a new instance of the field type.
    """
    instance = self.getContextInstance()
    fieldType = self.getFieldType()
    pvtName = self.getPrivateName()
    if isinstance(value, fieldType) or kwargs.get('_root', False):
      return setattr(instance, pvtName, value)
    setattr(instance, pvtName, fieldType(value))

  def __instance_delete__(self, *args, **kwargs) -> None:
    """
    Deletes the value of the field for the given instance. If the value is
    not set, it does nothing.
    """
    self._deletedGuard(self.getContextInstance(), self.__instance_get__())
    self.__instance_set__(DELETED, _root=True)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  Python API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __class_getitem__(cls, fieldType: type) -> Self:
    """
    Allows the AttriBox to be used as a generic type with a specified
    field type.
    """
    self = object.__new__(cls)
    self.__field_type__ = fieldType
    return self

  def __call__(self, *args: Any, **kwargs: Any) -> Any:
    """
    Allows the AttriBox to be called like a function, returning a new
    instance of the field type.
    """
    Object.__init__(self, *args, **kwargs)
    return self



"""QuestionableSyntax is raised when a name is encountered that is likely
a typo, such as '__set_item__' instead of '__setitem__' or '__setname__'
instead of '__set_name__'.
This is a"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class QuestionableSyntax(SyntaxError):
  """QuestionableSyntax is raised when a name is encountered that is likely
  a typo, such as '__set_item__' instead of '__setitem__' or '__setname__'
  instead of '__set_name__'.
  This is a subclass of SyntaxError and should be used to indicate that
  the code is likely to be incorrect. """

  __slots__ = ('derpName', 'realName',)

  def __init__(self, realName: str, derpName: str, ) -> None:
    self.derpName = derpName
    self.realName = realName
    SyntaxError.__init__(self, )

  def __str__(self) -> str:
    """String representation of the QuestionableSyntax."""
    spec = """%s! Received name '%s' which is similar enough to '%s' to be a 
    likely typo. """
    cls = type(self).__name__
    info = spec % (cls, self.derpName, self.realName)
    return textFmt(info)

  __repr__ = __str__



"""
TypeSig encapsulates type signatures and recognizes positional arguments
that match it.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from worktoy.waitaminute.dispatch import HashMismatch, TypeCastException
from worktoy.waitaminute.dispatch import CastMismatch, FlexMismatch
from worktoy.waitaminute import UnpackException
from worktoy.core.sentinels import THIS
from worktoy.utilities import unpack, bipartiteMatching
from . import PreClass, typeCast

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Iterator


class TypeSig:
  """
  #  DESCRIPTION
  'TypeSig' describes the types of an array of positional arguments
  analogous to how the 'type' function describes the 'type' of a single
  argument. The class supports both explicit types and the 'THIS'
  placeholder sentinel.

  #  NOMENCLATURE

  - 'Call arguments':  A tuple of objects. For example, the arguments
  received by a callable object.
  - 'type parsing':  The process by which the 'TypeSig' object determines
  how it best matches call arguments. If possible, the 'TypeSig' object
  will return a tuple of objects matching it. If not possible, a custom
  exception is raised. More details provided below.
  - 'type parser':  The 'TypeSig' instance method performing the 'type
  parsing'. The method receives 'call arguments' and returns 'processed
  arguments' (described below) or raises a custom exception specific to
  itself.
  - 'Processed arguments':  A tuple of objects returned by the 'TypeSig'
  object after parsing.
  - 'type casting':  same as 'type parsing' but emphasizes changes applied
  changing call arguments to processed arguments.

  #  USAGE
  #  #  Type Guarding and Casting
  A function requiring a very specific type signature, might use an
  instance of 'TypeSig' to type parse and possibly cast received arguments
  to match the required types. This allows the function more flexibility
  in what arguments it accepts while retaining the initial type
  specificity. Additionally, unsupported types are caught immediately and
  explicitly allowing for explicit and precise error handling.

  #  #  Hashable
  Since 'TypeSig' objects are hashable(*) they may be used as keys in
  'dict' objects. This facilitates dispatching of function calls to a type
  appropriate overload.

  #  IMPLEMENTATION DETAILS
  'TypeSig' provides multiple 'type parser' methods as listed below:

  - 'fast':  Strict, but fast hash-based parsing. This method will always
  return processed arguments identical to received call arguments when
  hashes match. Otherwise, it raises 'HashMismatch'.

  - 'cast'(**):  Applies type casting to each call argument not exactly
  matching the corresponding type. If any argument fails to cast,
  it raises 'CastMismatch' (a subclass of 'HashMismatch'). Otherwise,
  the type-casted arguments are passed to the 'fast' method described
  above to ensure type exactness. Please note that both this and 'flex'
  described below, are not appropriate for performance-sensitive
  applications, requiring substantial overhead compared to 'fast'.

  - 'flex'(**):  This method attempts to find a permutation of the call
  arguments that it might be able to cast. Additionally, if none of its
  raw types are 'list' or 'tuple', it will unpack any iterable call
  argument. If at all possible, the processed arguments will pass through
  the 'fast' method described above. Otherwise, it raises 'FlexMismatch' (
  a subclass of 'CastMismatch').

  (*) Note that 'TypeSig' objects are hashable only after the 'THIS'
  sentinel has been replaced with the appropriate class or a more specific
  hash-aware placeholder. 'worktoy' assigns to each class a hash value
  that depends only on names known ahead of actual class creation. This
  allows 'THIS' to be replaced with a 'PreClass' object having identical
  hash.

  (**) Note that 'TypeSig' parsing methods 'cast' and 'flex' do not
  support 'THIS', when used to overload the constructor of the class to
  which 'THIS' refers. The 'worktoy.static.Dispatch' class responsible for
  dispatching calls are meant to explicitly skip both 'cast' and 'flex'
  for this case. Since the 'TypeSig' object has no awareness of the
  broader context, it cannot detect and prevent this case. The 'cast' and
  'flex' methods will raise 'SignatureRecursion'

  #  ROLE IN LIBRARY
  The 'worktoy' library uses 'TypeSig' to provide function overloading.
  When the '@overload(...)' decorator receives types and decorates a
  function, it instantiates a 'TypeSig' object and associates the
  decorated function object with it. Please note that during class
  creation, methods created with 'def ...' begin as 'function' objects.
  They only become bound methods during the class creation process. Thus,
  any function decorator in Python is certain to receive a 'function
  object'. Even if this object later becomes a bound method.

  The 'worktoy' library leverages metaclass customization to facilitate
  function overloading. It uses the 'TypeSig' class when deciding which
  function object to dispatch. For more details, see the 'worktoy.mcls'
  documentation.

  # EXAMPLE
  from __future__ import annotations

  import sys
  from math import atan2

  def angle(*args, ) -> float:
    typeSig = TypeSig(float, float)  # Creates a float-float signature
    try:
      xp, yp = typeSig.fast(*args, )
    except HashMismatch as hashMismatch:
      pass
    else:
      if xp ** 2 + yp ** 2 > 1e-12:
        return atan2(yp, xp)
      raise ZeroDivisionError('Zero has no angle!')
    try:
      xp, yp = typeSig.cast(*args, )
    except CastMismatch as castMismatch:
      pass
    else:
      return angle(xp, yp)  # Recursive call
    try:
      xp, yp = typeSig.flex(*args, )
    except FlexMismatch as flexMismatch:
      raise
    else:
      return angle(xp, yp)  # Recursive call


  def main(*args) -> None:
    #  Main script entry point.
    try:
      res = angle(*args)
    except HashMismatch as hashMismatch:
      infoSpec = 'Unable to parse arguments: (%s), resulting in: %s'
      info = infoSpec % (str(args), hashMismatch)
      print(info)
      return 1
    except ZeroDivisionError as zeroDivisionError:
      infoSpec = 'Received origin point, resulting in: %s!'
      info = infoSpec % zeroDivisionError
      print(info)
      return 2
    except Exception as exception:
      infoSpec = 'Unexpected exception: %s'
      info = infoSpec % exception
      print(info)
      return 3
    else:
      infoSpec = 'Found angle: %.3f'
      info = infoSpec % res
      print(info)
      return 0
    finally:
      info = 'Exiting test of the TypeSig class!'
      print(info)


  if __name__ == '__main__':
    sys.exit(main(*sys.argv[1:]))
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __raw_types__ = None
  __type_casts__ = None
  __hash_value__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  CONSTRUCTORS   # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __init__(self, *types: Any) -> None:
    """Initialize the TypeSig instance."""
    self.__raw_types__ = types

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  PYTHON API   # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __iter__(self, ) -> Iterator[type]:
    """
    Implementation of iteration
    """
    yield from (self.__raw_types__ or ())

  def __hash__(self, ) -> int:
    """Forwards the hash to the hash of the types. If the replaceTHIS
    method has not been called, this will raise RuntimeError."""
    if self.__hash_value__ is None:
      return hash(self.__raw_types__)
    return self.__hash_value__

  def __eq__(self, other: object) -> bool:
    """Check if the other object is a TypeSig and has the same types."""
    if type(self) is not type(other):
      return NotImplemented
    if TYPE_CHECKING:  # pragma: no cover
      assert isinstance(other, TypeSig)
    for selfType, otherType in zip(self.__raw_types__, other.__raw_types__):
      if selfType is not otherType:
        return False
    return True

  def __len__(self, ) -> int:
    """Get the length of the types."""
    return len(self.__raw_types__)

  def __contains__(self, type_: type) -> bool:
    """Check if the type is in the types."""
    for raw in self.__raw_types__:
      if raw is type_:
        return True
    return False

  def __str__(self, ) -> str:
    """String representation reflects the types. """
    typeStr = ', '.join([t.__name__ for t in self.__raw_types__])
    info = """%s: [%s]""" % (type(self).__name__, typeStr)
    return info

  def __repr__(self, ) -> str:
    """Code representation reflects the types. """
    typeStr = ', '.join([t.__name__ for t in self.__raw_types__])
    info = """%s(%s)""" % (type(self).__name__, typeStr)
    return info

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def replaceTHIS(self, cls: type) -> None:
    """Replace the 'THIS' type with the class."""
    newTypes = []
    for type_ in self.__raw_types__:
      if type_ is THIS or isinstance(type_, PreClass):
        newTypes.append(cls)
        continue
      newTypes.append(type_)
    self.__raw_types__ = (*newTypes,)
    self.__hash_value__ = hash(self.__raw_types__)

  def fast(self, *args, ) -> tuple:
    """
    Hash-based type parsing. The method collects the types of the call
    arguments and compares the hash of the tuple containing them with the
    predicted hash. If hashes match, the call arguments are returned.
    Otherwise, raises 'HashMismatch'.

    This method is significantly faster than the 'cast' and 'flex'
    methods, but requires exact matching. Passing an 'int' object where a
    'float' object is expected, raises 'HashMismatch'.
    """
    if len(args) == len(self):
      if hash(self) == hash((*[type(arg) for arg in args],)):
        return (*args,)
    raise HashMismatch(self, *args)

  def cast(self, *args, **kwargs) -> tuple:
    """
    Iterates over the call arguments and the expected types:

    - If the argument at index 'i' is an instance of the type at 'i',
    the processed arguments will contain the argument at 'i'.
    - Otherwise, attempts to cast the argument to the type at 'i'. If
    successful, the processed arguments will contain the 'cast'
    argument at 'i'. If not, raises 'CastMismatch'.
    - If for every argument no 'CastMismatch' is raised, the
    processed arguments are returned as a tuple.
    """
    if len(args) != len(self):
      raise CastMismatch(self, *args)
    castArgs = []
    for type_, arg in zip(self.__raw_types__, args):
      if isinstance(arg, type_):
        castArgs.append(arg)
        continue
      try:
        castArg = typeCast(type_, arg)
      except TypeCastException as typeCastException:
        raise CastMismatch(self, *args) from typeCastException
      else:
        castArgs.append(castArg)
    else:
      return (*castArgs,)

  def flex(self, *args, **kwargs) -> tuple:
    """
    Attempts reordering of the call arguments to find a permutation
    that matches the types in the 'TypeSig'. If no permutation is found,
    raises 'FlexMismatch'. If a permutation is found, the processed
    arguments are returned as a tuple.
    """
    _ = hash(self)  # Ensure hash is set or raise RuntimeError immediately
    try:
      posArgs = unpack(*args, strict=True, shallow=True)
    except UnpackException as unpackException:
      pass
    else:
      return (*self.cast(*posArgs, **kwargs),)
    candidateMap = dict()
    candidateList = []
    for type_ in self:
      candidates = []
      for i, arg in enumerate(args):
        if isinstance(arg, type_):
          candidates.append(i)
      else:
        if not candidates:
          raise FlexMismatch(self, *args)
        candidateList.append((*candidates,))
    indices = bipartiteMatching(candidateList)
    if len(indices) != len(args):
      raise FlexMismatch(self, *args)
    return (*[args[i] for i in indices],)



"""
NameHook filters named used in the namespace system.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...waitaminute.meta import QuestionableSyntax, DelException
from . import AbstractSpaceHook

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, TypeAlias

  NearMiss: TypeAlias = tuple[str, str]


class NamespaceHook(AbstractSpaceHook):
  """
  NameHook intercepts names added to the namespace and filters out
  "near-miss" identifiers that resemble critical Python dunder methods.
  These mistakes often go unnoticed, leading to subtle bugs or broken
  protocol support.

  This hook raises a QuestionableSyntax exception when such names are
  detected during assignment in the namespace.

  ## Purpose

  Many magic methods in Python have specific names that must be spelled
  exactly. If a user misspells one by inserting or omitting underscores,
  the name is silently ignored by Python and treated as an ordinary
  attribute — sometimes shadowing a builtin or behaving unexpectedly.

  ## Near-miss Examples
  ___________________________________________________________________________
  | Intended Name  | Mistyped Name | Notes                                 |
  |----------------|---------------|---------------------------------------|
  | `__set_name__` | `__setname__` | Misses descriptor registration        |
  | `__getitem__`  | `__get_item__`| Breaks item access in dict-like APIs  |
  | `__setitem__`  | `__set_item__`| Same as above                         |
  | `__delitem__`  | `__del_item__`| Silent failure of delete protocol     |
  | `__delete__`   | `__del__`     | High risk: __del__ ties to GC hooks   |
  ¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨¨
  These errors can be especially difficult to diagnose, as they usually
  do not raise any errors directly — instead, they silently fail to
  participate in expected behaviors or override builtin methods.

  ## Usage
  To use NameHook, simply declare it in your namespace class:

  class Space(AbstractNamespace):  # Must inherit from AbstractNamespace
    #  Custom namespace class inheriting from AbstractNamespace
    nameHook = NameHook()  # Register the hook
"""

  @classmethod
  def _getClassDunders(cls) -> list[str]:
    """
    Get the class dunder names that are allowed in the namespace.
    """
    return [
        '__class_call__',
        '__class_init__',
        '__class_instancecheck__',
        '__class_subclasscheck__',
        '__class_str__',
        '__class_repr__',
        '__class_iter__',
        '__class_next__',
        '__class_bool__',
        '__class_contains__',
        '__class_len__',
        '__class_hash__',
        #  '__class_eq__',  Exhibits highly undefined behaviour!
        '__class_ne__',
        #  '__class_getitem__',  See footnote below
        '__class_setitem__',
        '__class_delitem__',
        '__class_getattr__',
        '__class_setattr__',
        '__class_delattr__',
        #  NOTE: '__class_getitem__' is commented out, rather than
        #  omitted, allowing this explanatory note to be found easily.
        #
        #  Since Python 3.7, support for the '__class_getitem__' became a
        #  feature of the interpreter itself. In fact, it inspired the
        #  implementation of all of the above class level hooks. Uniquely
        #  for '__class_getitem__', if it is defined on the class,
        #  the metaclass implementation of '__getitem__' will not be
        #  called.
        #
        #  This behaviour is opposite to the other hooks, as they
        #  are implemented here. 'AbstractMetaclass' provides
        #  implementations of the dunder methods, such that derived
        #  may implement the methods to achieve behaviour otherwise
        #  requiring metaclass reimplementation.
    ]

  @classmethod
  def _getNearMisses(cls) -> list[NearMiss]:
    """
    Get the near-miss names.
    """
    return [
        ('__set_name__', '__setname__'),  # NOQA, miss-spelled name
        ('__getitem__', '__get_item__'),
        ('__setitem__', '__set_item__'),
        ('__delitem__', '__del_item__'),
    ]

  @classmethod
  def _validateName(cls, name: str) -> bool:
    """
    Compares the name to list of potential near-miss names. If the name
    is a near-miss, a QuestionableSyntax exception is raised.
    """
    nearMisses = cls._getNearMisses()
    for nearMiss in nearMisses:
      if name == nearMiss[1]:
        raise QuestionableSyntax(*nearMiss, )
    return False

  def _validateDel(self, ) -> bool:
    """
    Validates that the current class is allowed to implement '__del__'.
    """
    if 'trustMeBro' in self.space.getKwargs():
      return True
    return False

  def setItemPhase(self, key: str, value: Any, oldValue: Any) -> bool:
    """
    Hook for setItem. This is called before the __setitem__ method of
    the namespace object is called. The default implementation does nothing
    and returns False.
    """
    if key == '__del__':
      if self._validateDel():
        return False
      mcls = self.space.getMetaclass()
      name = self.space.getClassName()
      bases = self.space.getBases()
      raise DelException(mcls, name, bases, self.space)
    return self._validateName(key)

  def postCompilePhase(self, compiledSpace: dict) -> dict:
    """
    Hook for postCompile. This is called after the __init__ method of
    the namespace object is called. The default implementation does nothing
    and returns the contents unchanged.
    """
    dunderNames = self._getClassDunders()
    for name in dunderNames:
      try:
        _ = self.space.deepGetItem(name)
      except KeyError as keyError:
        compiledSpace[name] = self._defaultDunderFactory(name)
        continue
      else:
        continue
    return compiledSpace

  @staticmethod
  def _defaultDunderFactory(name: str) -> classmethod:
    """
    Default dunder factory for the namespace hook.
    """

    def dunderStruck(cls, *args, **kwargs) -> str:
      """
      Default dunder method for the namespace hook.
      """
      raise NotImplementedError

    setattr(dunderStruck, '__name__', name)
    setattr(dunderStruck, '__qualname__', name)
    setattr(dunderStruck, '__dunder_struck__', True)

    return classmethod(dunderStruck)



"""SubclassException should be raised when a class is not a subclass of
the expected base class. """
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class SubclassException(TypeError):
  """SubclassException should be raised when a class is not a subclass of
  the expected base class."""

  __slots__ = ('cls', 'expected')

  def __init__(self, cls: type, base: type) -> None:
    """Initialize the exception with the object and expected base class."""
    self.cls, self.expected = cls, base
    TypeError.__init__(self, )

  def __str__(self) -> str:
    """
    Return a string representation of the SubclassException object.
    """
    infoSpec = """Expected class '%s' to be a subclass of '%s'!"""
    clsName = type(self.cls).__name__
    baseName = type(self.expected).__name__
    info = infoSpec % (clsName, baseName)
    return textFmt(info)

  __repr__ = __str__



"""
Code example from 'TypeSig' documentation.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

import sys
from math import atan2

from worktoy.waitaminute import HashMismatch, CastMismatch, FlexMismatch
from worktoy.static import TypeSig


def angle(*args, ) -> float:
  """
  Computes the argument of the complex number: x + yJ
  """
  typeSig = TypeSig(float, float)  # Creates a float-float signature
  try:
    xp, yp = typeSig.fast(*args, )
  except HashMismatch as hashMismatch:
    pass
  else:
    if xp ** 2 + yp ** 2 > 1e-12:
      return atan2(yp, xp)
    raise ZeroDivisionError('Zero has no angle!')
  try:
    xp, yp = typeSig.cast(*args, )
  except CastMismatch as castMismatch:
    pass
  else:
    return angle(xp, yp)  # Recursive call
  try:
    xp, yp = typeSig.flex(*args, )
  except FlexMismatch as flexMismatch:
    raise
  else:
    return angle(xp, yp)  # Recursive call


def main(*args) -> int:  # NOQA: Pycharm, read up on 'finally'!
  """Main script entry point."""
  try:
    res = angle(*args)
  except HashMismatch as hashMismatch:
    infoSpec = 'Unable to parse arguments: (%s), resulting in: %s'
    info = infoSpec % (str(args), hashMismatch)
    print(info)
    return 1
  except ZeroDivisionError as zeroDivisionError:
    infoSpec = 'Received origin point, resulting in: %s!'
    info = infoSpec % zeroDivisionError
    print(info)
    return 2
  except Exception as exception:
    infoSpec = 'Unexpected exception: %s'
    info = infoSpec % exception
    print(info)
    return 3
  else:
    infoSpec = 'Found angle: %.3f'
    info = infoSpec % res
    print(info)
    return 0
  finally:
    info = 'Exiting test of the TypeSig class!'
    print(info)


if __name__ == '__main__':
  sys.exit(main(*sys.argv[1:]))



"""
'PreClassHook' replaces 'THIS' in the AbstractNamespace with 'PreClass'
objects having the hash and name of the future class ahead of class creation.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...static import PreClass, TypeSig
from ...waitaminute import TypeException
from . import AbstractSpaceHook

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any


class PreClassSpaceHook(AbstractSpaceHook):
  """
  PreClassHook resolves the circular reference problem caused when a
  decorator or hook needs to refer to the class under construction —
  *before* that class exists.

  ## Purpose

  During class construction, decorators (such as those involved in
  overload resolution) may need to reference the class currently being
  defined. But this isn't possible with a literal reference, since the
  class object doesn’t exist yet.

  To solve this, a sentinel object named `THIS` is used as a placeholder.
  The PreClassHook intercepts assignments in the namespace and replaces
  any references to `THIS` found in function type signatures with a
  special `PreClass` proxy.

  The `PreClass` object mimics the identity of the future class:
  - It has the same `__hash__` as the future class will
  - It stores the intended class name and metaclass
  - It can be used in hash-based fast dispatch without knowing the class

  ## Behavior

  - **During setItemHook**:
    If a function assigned to the namespace contains one or more
    `TypeSig` objects (e.g. for overloads), each is scanned for
    references to `THIS`. These are replaced in-place with the
    appropriate `PreClass` object.

  - **During postCompileHook**:
    After the class body has executed, the overload map is scanned for
    type signatures that refer to known base classes. For each such
    signature, a new equivalent signature is created where the base class
    is replaced with the `PreClass` object. This allows overloads in
    parent classes referencing `THIS` to resolve correctly in subclasses.

  ## Usage

  This hook is registered inside a namespace like so:

      class MyNamespace(AbstractNamespace):
        preClassHook = PreClassHook()

  It cooperates with the OverloadHook to ensure that decorators using
  `THIS` for self-type references work reliably across complex inheritance
  hierarchies.

  ## Notes

  - The PreClass object is cached internally and created on demand.
  - It is only instantiated once per class construction.
  - It is safe to hash and compare, but is not a fully functional class.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private variables
  __pre_class__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _createPreClass(self, **kwargs) -> None:
    """
    Creates the unique 'PreClass' object for the namespace instance owning
    this hook instance.
    """
    _hash = self.space.getHash()
    _name = self.space.getClassName()
    _meta = self.space.getMetaclass()
    self.__pre_class__ = PreClass(_hash, _name, _meta)

  def _getPreClass(self, **kwargs) -> PreClass:
    """
    Getter-function for the 'PreClass' for the namespace instance owning
    this hook instance.
    """
    if self.__pre_class__ is None:
      if kwargs.get('_recursion', False):
        raise RecursionError  # pragma: no cover
      self._createPreClass()
      return self._getPreClass(_recursion=True)
    if isinstance(self.__pre_class__, PreClass):
      return self.__pre_class__
    raise TypeException('__pre_class__', self.__pre_class__, PreClass, )

  def setItemPhase(self, key: str, val: Any, old: Any, ) -> bool:
    """
    If key contains reference the class under construction by containing
    'THIS', replace with 'PreClass' object providing the hash and name of the
    future class.
    """
    preClass = self._getPreClass()
    typeSigs = getattr(val, '__type_sigs__', None)
    if typeSigs is None:
      return False
    for sig in typeSigs:
      if isinstance(sig, TypeSig):
        TypeSig.replaceTHIS(sig, preClass, )
        _ = hash(sig)
        continue
      raise TypeException('sig', sig, TypeSig, )
    else:
      return False

  def postCompilePhase(self, compiledSpace) -> dict:
    """
    Where a type signature in the overload map includes a base class,
    we create a new type signature pointing to the same function, but with
    the 'PreClass' object replacing the base class. This will allow 'THIS'
    defined in an overload in a parent to work in the child class with
    either an instance of the parent or the child.
    """
    overloadMap = self.space.getOverloadMap()  # str: dict[TypeSig, Func]
    preClass = self._getPreClass()
    bases = self.space.getBases()
    newLoadMap = dict()
    for key, sigMap in overloadMap.items():
      moreSigMap = dict()
      for sig, func in sigMap.items():
        newSig = []
        for type_ in sig:
          if type_ in bases:
            newSig.append(preClass)
            continue
          newSig.append(type_)
        else:
          if preClass in newSig:
            newTypeSig = TypeSig(*newSig)
            moreSigMap[newTypeSig] = func
          moreSigMap[sig] = func
      else:
        newLoadMap[key] = moreSigMap
    else:
      self.space.__overload_map__ = newLoadMap
    return compiledSpace



"""
OWNER provides a sentinel object used together with THIS by the descriptor
flow where OWNER provides a placeholder for the owning class, while THIS
refers to an instance of the class (self typically).
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Sentinel

if TYPE_CHECKING:  # pragma: no cover
  pass


class OWNER(Sentinel):
  """
  OWNER provides a sentinel object used together with THIS by the descriptor
  flow where OWNER provides a placeholder for the owning class, while THIS
  refers to an instance of the class (self typically).
  """



"""
The 'worktoy.keenum' module provides the enumerating KeeNum class.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from ._keenum_base import _KeeNumBase  # Private
from ._auto import auto
from ._auto import _AutoMember
from ._kee_hook import KeeSpaceHook
from ._kee_space import KeeSpace
from ._kee_meta import KeeMeta
from ._keenum import KeeNum

__all__ = [
    'auto',
    'KeeSpaceHook',
    'KeeSpace',
    'KeeMeta',
    'KeeNum',
]



"""
Field provides a property-like descriptor implementation allowing
descriptor owners to decorate methods to designate them as accessors.
Since these are identified by name, the function object are entirely
unaffected by the decoration and subclasses can override any decorated
method and the descriptor uses the overridden method instead.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..core import Object
from ..utilities import maybe
from ..waitaminute import TypeException
from ..waitaminute.desc import ProtectedError, ReadOnlyError, AccessError

if TYPE_CHECKING:  # pragma: no cover
  from typing import Any, Callable


class Field(Object):
  """
  Flexible descriptor requiring accessor methods to be decorated. Please
  note that the instance of 'Field' can decorate only methods appearing
  below it in the class body.

  @GET - Decorate one method designating it as the 'getter'. It should be
  a normal instance method that can be run without any other arguments.

  @SET - Decorate any number of methods as setters. Every such method runs
  in response to __set__

  @DEL - Decorate any number of methods as deleters. Optionally, implement
  by setting the value to the 'DELETED' sentinel object.
  """

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  NAMESPACE  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  #  Private Variables
  __get_key__ = None
  __set_keys__ = None
  __delete_keys__ = None

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  GETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def _getGetterKey(self) -> str:
    if isinstance(self.__get_key__, str):
      return self.__get_key__
    if self.__get_key__ is None:
      raise AccessError(self)
    raise TypeException('__get_key__', self.__get_key__, str)

  def _getSetterKeys(self, newValue: Any = None) -> tuple[str, ...]:
    """Iterates over setter keys. The 'newValue' is used only to compose
    error message. """
    return (*[k for k in maybe(self.__set_keys__, ()) if k],)

  def _getDeleterKeys(self) -> tuple[str, ...]:
    """
    Returns the keys of the deleter methods. If no deleter methods are
    registered, returns an empty tuple.
    """
    return (*[k for k in maybe(self.__delete_keys__, ()) if k],)

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  SETTERS  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def GET(self, callMeMaybe: Callable) -> Callable:
    """
    Decorator for the getter method. The method should be a normal instance
    method that can be run without any other arguments.
    """
    self.__get_key__ = callMeMaybe.__name__
    return callMeMaybe

  def SET(self, callMeMaybe: Callable) -> Callable:
    """
    Decorator for the setter method. The method should be a normal instance
    method that can be run without any other arguments.
    """
    existing = maybe(self.__set_keys__, ())
    self.__set_keys__ = (*existing, callMeMaybe.__name__,)
    return callMeMaybe

  def DELETE(self, callMeMaybe: Callable) -> Callable:
    """
    Decorator for the deleter method. The method should be a normal instance
    method that can be run without any other arguments.
    """
    existing = maybe(self.__delete_keys__, ())
    self.__delete_keys__ = (*existing, callMeMaybe.__name__,)
    return callMeMaybe

  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  #  DOMAIN SPECIFIC  # # # # # # # # # # # # # # # # # # # # # # # # # # # #
  # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #

  def __instance_get__(self, *args, **kwargs) -> Any:
    """
    Retrieves the getter from the owner of instance and the registered
    name of getter function. Please note that while the instance received
    is certain to satisfy: isinstance(instance, self.getFieldOwner()),
    the field owner is the class where the descriptor was instantiated.
    For this reason, the decorated method is retrieved by name from the
    owner of the instance received.
    """
    getterKey = self._getGetterKey()
    getterFunc = getattr(self.getContextInstance(), getterKey, )
    return getterFunc(*args, **kwargs)

  def __instance_set__(self, val: Any, *args, **kwargs) -> None:
    """
    All decorated setters are retrieved in the same fashion as the getter.
    """
    setterKeys = self._getSetterKeys(val)
    if not setterKeys:
      raise ReadOnlyError(self.instance, self, val, )
    instance = self.getContextInstance()
    setterFuncs = [getattr(instance, k) for k in setterKeys]
    setterFuncs = [f for f in setterFuncs if callable(f)]
    for setterFunc in setterFuncs:
      setterFunc(val, *args, **kwargs)

  def __instance_delete__(self, *args, **kwargs) -> None:
    """
    All decorated deleters are retrieved in the same fashion as the getter.
    """
    deleterKeys = self._getDeleterKeys()
    if not deleterKeys:
      raise ProtectedError(self.instance, self, )
    for key in deleterKeys:
      getattr(self.instance, key, )(**kwargs)



"""
The metaclass 'KeeMeta' automatically creates the KeeNum class for which
the class here is a placeholder of sorts.

"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import KeeMeta

if TYPE_CHECKING:  # pragma: no cover
  from typing import Iterator, Any


  class Member:
    """
    Is that enough, pycharm?
    """

    name: str
    value: Any


  class KeeNum(dict):
    """
    This is just so pycharm will chill out with warnings about KeeNum
    """

    name: str
    value: Any

    def __init__(self, *args, **kwargs) -> None:
      """KeeNum is a placeholder for the KeeMeta metaclass."""

    def __iter__(self) -> Iterator:
      pass

    def __getitem__(self, item) -> Member:
      pass

    def __setitem__(self, key, value) -> None:
      pass
else:
  KeeNum = KeeMeta.getCoreKeeNum()



"""
DESC sentinel provides a placeholder for the descriptor object in the
descriptor flow along with THIS and OWNER.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import Sentinel

if TYPE_CHECKING:  # pragma: no cover
  pass


class DESC(Sentinel):
  """
  DESC sentinel provides a placeholder for the descriptor object in the
  descriptor flow along with THIS and OWNER.
  """



"""EZMeta provides the metaclass for the EZData class."""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ..mcls import AbstractMetaclass, Base
from ..ezdata import EZSpace

if TYPE_CHECKING:  # pragma: no cover
  pass


class EZMeta(AbstractMetaclass):
  """EZMeta provides the metaclass for the EZData class."""

  @classmethod
  def __prepare__(mcls, name: str, bases: Base, **kwargs: dict) -> EZSpace:
    """Prepare the class namespace."""
    return EZSpace(mcls, name, bases, **kwargs)



"""
MetaObject provides the metaclass for the 'Object' class in the
'worktoy.core' package. It derives from MetaType, allowing subclasses of
derived classes to use other metaclass derived from MetaType.

Please note the difference between a metaclass and a baseclass. To
distinguish, this library proposes the following terminology:

 - A class 'derives from' the metaclass creating it. Ultimately,
 all derives from 'type', possibly with more metaclasses in between. The
 'worktoy' library provides 'MetaType' as the initial metaclass.

 - A class 'is based on' the baseclasses from which it inherits.
 Ultimately, all classes are based on 'object', possibly with more classes
 in between. The 'worktoy' library provides 'Object' as the initial
 baseClass.

with the following nomenclature:

 - class: (Traditional definition)
 - metaclass: A class whose instances are themselves classes.
 - meta-metaclass: A class whose instances are themselves metaclasses.
 - derive: The creation of a new class from a metaclass.
 - base: The extension of a class from a baseclass.

No distinction is necessary when class definition is considered
an implementation detail across object orientated programming
languages. Instead, the computer science field has extensively studied the
relationships between classes and baseclasses. By elevating classes to
first class citizens, Python introduces a concept semantically similar to
class inheritance, but fundamentally different.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from . import MetaType

if TYPE_CHECKING:  # pragma: no cover
  pass


class MetaObject(MetaType, metaclass=MetaType):
  """
  MetaObject provides the metaclass for the 'Object' class in the
  'worktoy.core' package. It derives from MetaType, meaning that a
  subclass of 'Object' can still use a custom metaclass, provided that
  metaclass also derives from 'MetaType'.
  """
  pass



"""
EmptyKeeNumError is a custom exception class raised to indicate that a
KeeNum class failed to provide any enumeration members.
"""
#  AGPL-3.0 license
#  Copyright (c) 2025 Asger Jon Vistisen
from __future__ import annotations

from typing import TYPE_CHECKING

from ...utilities import textFmt

if TYPE_CHECKING:  # pragma: no cover
  pass


class EmptyKeeNumError(Exception):
  """
  EmptyKeeNumError is a custom exception class raised to indicate that a
  KeeNum class failed to provide any enumeration members.
  """

  __slots__ = ('className',)

  def __init__(self, className: str) -> None:
    """Initialize the EmptyKeeNumError object."""
    self.className = className
    Exception.__init__(self, )

  def __str__(self, ) -> str:
    """Return the string representation of the EmptyKeeNumError object."""
    infoSpec = """KeeNum class '%s' has no members!"""
    info = infoSpec % self.className
    return textFmt(info)

  __repr__ = __str__
